%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
%-------------------- Quantitative Microwave Holography -------------------
%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
% Written by Daniel Tajik, McMaster University
% This code is confidential and cannot be used for purposes beyond
% educational study.
%--------------------------------------------------------------------------
%
% Written to complement:
%   Introduction to Microwave Imaging, by Natalia K. Nikolva
%   Section 5.2: 3D holographic inversion with wideband planar data acq.
%
% 
% Original Version: 1 (July 5th, 2017)
% Revision: 2 (July 5th, 2018)
% Changes:
% - Added optimization techniques to remove nonphysical values
% - Changed variable names to improve readability/comparability to papers
% - Include a simple application of pseudoinverse to improve image
% reconstruction
% - Modified code to improve reconstruction speed on large number of
% datasets
% Revision: 3(March 23, 2021)
% - Added a combined Born/Rytov method
% - Implemented a basic Matrixless approach (under development)
% - Added a filter correction factor to improve permittivity accuracy
% - Added Rytov-Real approach
% - Added Preprocessing visualization for improved optimization of filters
% Revision: 4(June 23, 2021)
% - Improved commenting
% - Added parameter display at end of execution
% - Improved low-pass filter alpha scaling
% - Adjusted  sizes of figures, figure position
%--------------------------------------------------------------------------


% Clear all previous information every execution, and calculate runtime.
clc;
clear all;
close all;
tic

%% Initial Information: Loading data files, declaring variables
%--------------------------------------------------------------------------
%---------------------------USER INPUT BELOW-------------------------------
%-------------------------LOAD YOUR DATA HERE------------------------------
%--------------------------------------------------------------------------
%-------------------------SIMULATION EXAMPLE-------------------------------
%--------------------------------------------------------------------------

% EXAMPLE
% This is the data used to generate the images from the experiment
% performed in [Tajik, EURAD 2019]. More information can be found here: 
% https://ieeexplore-ieee-org.libaccess.lib.mcmaster.ca/document/8904477

% Begin loading the files containing the RO (reference object), CO
% (calibration object) and OUT (object-under-test) datasets. All datasets
% are acquired using S-parameters.
% Here, S31 = Transmission measurements at off-angle bowtie antennas.
% Naming is due to the RF switch that was used at the time. There were 4
% antennas measured through port 3, labelled a,b,c,d.
%       S41 = Transmission coefficient between horn and central bowtie
%       element
%
%--------------------------------------------------------------------------
% LOAD AND ORGANIZE DATASETS
%--------------------------------------------------------------------------

%We can narrow our field of view or our frequency range if required
xrng = 1:48;
yrng = 1:48;
rng = 1:51;

%Measurements of the OUT and RO
load S31_OUT_abcd.mat
load S41_OUT.mat
load S31_RO_abcd.mat
load S41_RO.mat

%Load the cell structures that are used to simplify arrays
s_RO = {S31_RO_abcd(xrng,yrng,rng,1);S31_RO_abcd(xrng,yrng,rng,2);S31_RO_abcd(xrng,yrng,rng,3);S31_RO_abcd(xrng,yrng,rng,4);S41_RO(xrng,yrng,rng)};
s_OUT = {S31_OUT_abcd(xrng,yrng,rng,1);S31_OUT_abcd(xrng,yrng,rng,2);S31_OUT_abcd(xrng,yrng,rng,3);S31_OUT_abcd(xrng,yrng,rng,4);S41_OUT(xrng,yrng,rng)};

%Here, a separate RO was used for the CO due to the significant background
%artifacts cluttering the CO measurement. It leads to an improved
%reconstruction

load S31_CO_abcd_abs.mat
load S41_CO_abs.mat
load S31_RO_abcd_abs.mat
load S41_RO_Abs.mat

%Load the cell structures that are used to simplify arrays
s_RO_forCO = {S31_RO_abcd(xrng,yrng,rng,1);S31_RO_abcd(xrng,yrng,rng,2);S31_RO_abcd(xrng,yrng,rng,3);S31_RO_abcd(xrng,yrng,rng,4);S41_RO(xrng,yrng,rng)};
s_CO = {S31_CO_abcd(xrng,yrng,rng,1);S31_CO_abcd(xrng,yrng,rng,2);S31_CO_abcd(xrng,yrng,rng,3);S31_CO_abcd(xrng,yrng,rng,4);S41_CO(xrng,yrng,rng)};




%--------------------------------------------------------------------------
% DECLARE VARIABLES FOR QUANTITATIVE RECONSTRUCTION
%--------------------------------------------------------------------------
% (UNIT OF MEASUREMENT = METERS)

% Relative permittivity of background medium
eps_RO = 11-5i;

% Relative permittivity of the scattering probe
eps_CO = 50;

% Volume of the scattering probe 
vol_sp = pi*(5e-3)^2*10e-3; %cylinder, r = 5 mm, h = 10 mm 

% Scan step
xstep = 3e-3;

%Range (depth) resolution
z= 55e-3;

%Frequency table
ftable = 3e9:1e8:8e9;

%--------------------------------------------------------------------------
% ADDITIONAL SETTINGS 
%--------------------------------------------------------------------------
% Replace 0 with 1 to enable the boolean variables

% The default approximation used in this code is Born's approximation (see
% section 1.14). However, this method can also operate using Rytov's
% approximation (see section 1.16). Setting the parameter to 1 will execute
% the code using Rytov's approximation. 
% Note that when only computing Rytov, the code will not perform phase
% unwrapping, and thus in real-world measurements, the reconstruction
% quality can be significantly poorer.
enable_rytov = 0;

% If you are aware that there is significant phase wrapping in your signal,
% enable this over the above function. It will compute the reconstruction
% using only the real component of the OUT. This is the same approach used
% in generating the signals for the combined-Born/Rytov approach
enable_rytov_real = 0;

% If enabled, will compute the combined Born/Rytov approximation. This
% method was briefly discussed in:
% https://ieeexplore-ieee-org.libaccess.lib.mcmaster.ca/document/8904477
% enabling this over-rides other settings.
comb_bornrytov = 1;

% Removes the background RO from the CO and OUT data. If Rytov is selected,
% it will perform the Rytov version of de-embedding (see [Tajik, JPIER 2017])
subtract_RO = ones(length(s_CO));

% PSF Clean
% Use Jessica Trac's algorithm to remove the PSFs that are low quality. This
% approach was discussed in:
% https://ieeexplore-ieee-org.libaccess.lib.mcmaster.ca/document/8747500
clean_psf = 0;

% Apodization filtering: This filter will alleviate errors related to the
% DFT as shown in the work of [Tajik, JPIER 2017]. While it has no
% impact on simulated data, it is typically required when processing
% experimental data.
apod = 1;
% Shows the apodization filter currently used.
show_apod = 1;

% Post-inversion low-pass filtering: This filter will alleviate errors
% related to oversampling (see section 5.3). In the simulation example 
% provided, this is not the case, and is not necessary. In experimental 
% data, this filter is a necessity.
enable_lpf = 1;
% Scale the size of the filter to improve optimization (between 0 and 1)
filter_scale = 0.4;
% Shows the apodization filter.
show_lpf = 1;

% Filtering can be applied to focus in on particular areas as discussed in
% [Tajik, JPIER 2017].
focusing = 0;
focus_center = [24,24];
focus_size = 30;
show_focus = 0;

%Smoothing filter: This increases the sampling rate of the final image, and
%smooths it with a low pass filter. Use if you would like to see a smoother
%result.
smoothing = 0;

% Save the images if desired
save_comb = 0;
save_ind = 0;

% The images can either be generated per each S-parameter, or all together.
% The simulation requires all 4 S-parameters.
close_comb = 0;
close_ind = 1;


%--------------------------------------------------------------------------
% SETTINGS UNDER DEVELOPMENT - USE AT OWN RISK 
%--------------------------------------------------------------------------
% Execute Matrixless QMH
matrixless = 0;

% Use the speherical wave approximation in place of a measurement PSF
edot = 0;

% Preprocessing generates images of the summed signals to allow for visual
% interpretation of regions of interest. It also highlight strong responses
% induced by air gaps which can be damaging on the signal. Consider
% optimizing the apodization filter if the airgaps are deep within the
% imaged domain.
preprocess = 0;

% go through the image and identify regions that are outside of 10% of the
% permittivies and set everything else to be 0.
combine_COs = 0;

% Goes beyond the Low-pass filter, filtering based on each frequency.
freq_filter = 0;

% Sets the tolerance for the pinv function based on the
% map of SVD values. Not necessary in simulation reconstructions. This
% strategy may replace the two previous filters when optimized.
set_tol = 0;

show_svd = 0;
% Removing nonphysical values: Enable this if you wish to remove
% nonphysical values in the final reconstruction. NOTE: YOU MUST HAVE CVX
% INSTALLED AND LINKED TO MATLAB IN ORDER FOR THIS FUNCTION TO WORK! You
% can find a guide to do this at cvxr.com
correct_nonphys = 0;

%--------------------------------------------------------------------------
% PARAMETER DELCARATIONS 
%--------------------------------------------------------------------------
% Sets appropriate parameters used later on in the code. No adjustment 
% necessary unless specifically required.

% declaring required parameters used further on in the code.
ns = size(s_CO,1); 
[nx ny nf nz] = size(s_CO{1});
c0 = 2.998e8;
cb = c0/abs(eps_RO);
xrng = xstep*(nx-1);
x=linspace(-xrng/2,xrng/2,nx);
y=linspace(-xrng/2,xrng/2,ny);
% if it is a 2D image, zstep is the distance between antennas.
if length(z)>1
    zstep = z(2)-z(1);
else
    zstep = z(1);
end
units = 'm';

%override settings to ensure successful reconstruction.

if enable_rytov_real && ~comb_bornrytov
    enable_rytov = 0;
    disp('Computing Rytov-Real only Approach')
elseif comb_bornrytov
    enable_rytov = 0;
    enable_rytov_real = 1;
    disp('Computing Combined Born/Rytov Approach')
elseif enable_rytov
    disp('Computing Rytov Approach - Be Aware of Phase Wrapping Issues!')
else
        disp('Computing Born Approach')
end



%% UNDER DEVELOPMENT: Generate the dot product for E-field.
% only setup for a particular experiment, will not work with any data

% x is up(-) down(+)
% y is left(-) right(+)
if edot
z = 0.03
for ss = 1:ns
    if ss == 1
        ynew{1} = y - 10e-3; %right
        xnew{1} = x + 10e-3; %up
    elseif ss == 2
        ynew{2} = y + 10e-3; %left
        xnew{2} = x + 10e-3; %up
    elseif ss == 3
        ynew{3} = y + 10e-3; %left
        xnew{3} = x - 10e-3; %down
    elseif ss == 4
        ynew{4} = y - 10e-3; %right
        xnew{4} = x - 10e-3; %down
    else
        ynew{5} = y;
        xnew{5} = x;
    end
    
    for ii = 1:nx
        for jj = 1:ny
            for kk = 1:nz
                r_vec{ss}(ii,jj,kk) = sqrt(xnew{ss}(ii)^2 + ynew{ss}(jj)^2 + z(kk)^2);
            end
        end
    end
end

for ss = 1:ns

for ff = 1:nf  
    lambda = c0/ftable(ff)*sqrt(abs(eps_RO));
    
    k_0 = 2*pi/lambda;
    res_kern(:,:,:,ff) = exp(-1i*k_0*r_vec{ss})./r_vec{ss} .* exp(+1i*k_0*r_vec{ss})./r_vec{ss};
end

    S_res{ss} = res_kern(:,:,:,:)./10000;
end
end

%--------------------------------------------------------------------------
%---------------------------END OF USER INPUT------------------------------
%------------TREAD CAUTIOUSLY WHEN EDITTING BEYOND THIS POINT--------------
%--------------------------------------------------------------------------

%% Stage One: Quantitative Microwave Holography
%--------------------------------------------------------------------------
%The basics of this algorithm can be seen in Section 5.2 of Introduction to 
%Microwave Imaging. It has the following stages:
%1.1 Acquisition of scattering parameter data
%1.2 Apodization Filtering
%1.3 Normalization and the Fast Fourier Transform 
%1.4 Inversion step at each spectral position k
%1.5 Post inversion low pass filtering
%1.6 Multiplicative factoring and final image adjustments
%1.7 Image Generation
%--------------------------------------------------------------------------

%% Part 1.1: Acquiring the scattered response (Born or Rytov Approximation)
%
% Approximate the incident field if required.
% The scattered response is derived from the CO and OUT responses

for i=1:length(s_RO) 
      
    if size(s_RO{i},1) == 1
        for f = 1:nf
            % in our simulation data, the RO is typically one measurement
            % and needs to be built into a map, which is not the case for
            % experimental measurements
            grad = 0:1/(nx-1):1;
            thresh = 0*exp(1i*0.8);
            s_RO_hold{i}(:,:,f) = s_RO{i}(1,1,f)*ones(nx,ny)+thresh*(grad'*grad)*s_RO{i}(1,1,f);
        end
        s_RO{i} = s_RO_hold{i};
    end
       % You can use separate ROs for the CO vs OUT measurement, by default is not
% applied
end

if exist('s_RO_forCO','var') == 0
   disp('The RO dataset is the same for both the CO and OUT');
   s_RO_forCO = s_RO;
else
   disp('The RO dataset is independent for both the CO and OUT');
end

%BORN STRATEGY
%----------------------------------------------------------------------
for i=1:length(s_RO) 

    % Scattered response is S_SC = S_(OUT,CO) - S_RO
    if subtract_RO(i) == 1 && ~enable_rytov
        %loop through each frequency
        for f = 1:nf
            sc_OUT{i}(:,:,f) = s_OUT{i}(:,:,f) - s_RO{i}(:,:,f,1);
            
            %loop through each z-plane
            for iz = 1:nz
                sc_CO{i}(:,:,f,iz) = s_CO{i}(:,:,f,iz) - s_RO_forCO{i}(:,:,f);
                
            end
        end
    end
end

%RYTOV STRATEGY
%----------------------------------------------------------------------
% Scattered response is S_SC = S_RO*ln(S_(OUT,CO)/S_RO)
% Be aware that phase wrapping leads to large artifacts being generated
% in the reconstruction. Removal of phase wrapping is challenging if
% the wrapping is occuring in the spatial domain (rather than the
% frequency domain as is typically understood).
%

if enable_rytov==1
    if isempty(s_RO)
        disp('No RO was loaded! Output will be incorrect.')
        enable_rytov = 0;
    else
        %loop through each s parameter
        for n = 1:ns
            sc_OUT{n} = s_RO{n}.*(log(s_OUT{n}./s_RO{n}));
 
            %loop through each z plane
            for iz = 1:nz 
                sc_CO{n}(:,:,:,iz) =  s_RO{n}(:,:,:,1).*(log(squeeze(s_CO{n}(:,:,:,iz))./s_RO_forCO{n}(:,:,:,1)));
            end          
        end
    end
end 

rng_goodphase = 1:nf;


% NEW: Combined Born and Rytov Strategy
%----------------------------------------------------------------------
% We can combine both the Born and Rytov techniques by breaking down the
% methods into their real and imaginary components, and solving them all
% together. Note that we modify Rytov in order to avoid phase wrapping.

if enable_rytov_real ==1% && enable_rytov ==0;
    if isempty(s_RO)
        disp('No RO was loaded! Output will be incorrect.')
        enable_rytov = 0;
    else
        disp('Generating the separate Real/Imaginary components for reconstruction');
        
        %If we are solving the combination of Born and Rytov, we already
        %have the Born solution. We need to redefine how to determine Rytov
        %in order to extract the correct responses. To avoid the issue of
        %Phase in Rytov, we extract the magnitude components
                

        for n = 1:ns
            s_OUT_ryt{n} = abs(s_RO{n}).*(log(abs(s_OUT{n})./abs(s_RO{n})));
 
            %loop through each z plane
            for iz = 1:nz 
                
                %Interestingly, in some situations for simulated data
                %-infinity
                %data can be generated in the CO measurement. For now equate this
                %to the minimum value.
                check_zeros = squeeze(s_CO{n}(:,:,:,iz));
                min_nonzero = min(abs(check_zeros(check_zeros>0)));
                check_zeros(check_zeros==0) = min_nonzero;
                
                %Obtain the real part of the CO, set all infinites to min
                CO_real = log(abs(s_CO{n}(:,:,:,iz))./abs(s_RO_forCO{n}(:,:,:)));
                CO_real(isinf(CO_real)) = min_nonzero;
                
                % note, there are better 2D phase unwrapping approaches. 
                % take a look at
                % https://www.mathworks.com/matlabcentral/fileexchange/65565-fast-2d-phase-unwrapping
                CO_imag_unwrap = unwrap(unwrap(angle(s_CO{n}(:,:,:,iz)),[],1),[],2) - unwrap(unwrap(angle((s_RO_forCO{n}(:,:,:))),[],1),[],2);
                
                %used later when cleaning up the reconstruction
                CO_imagchk{n}(:,:,:,iz) = CO_imag_unwrap;
                
                %generate the Rytov based approach.
                s_CO_ryt{n}(:,:,:,iz) = abs(s_RO_forCO{n}(:,:,:,1)).*(CO_real + 1i*CO_imag_unwrap);
                
                
                if isinf(s_CO_ryt{n}(:,:,:,iz))
                    s_CO_ryt{n}(isinf(s_CO_ryt{n}(:,:,:,iz))) = log(min_nonzero./abs(s_RO_forCO{n}(:,:,:)));
                    
                end
                
            end
        end
        
    end

    disp("Parameters Loaded and Scattered Parameters Generated")
toc
    
%Range Cleanup
% When we perform Rytov, sometimes we see phase wrapping that cannot be
% corrected. In these cases, we can simply throw out these frequencies. We
% can find which PSFs are affected by finding the gradient of the phase of
% the signal (in this case, the gradient of the signal will show a
% derivative greater than pi/2).
%ff_set = [1:51;1:51;1:51;1:51;1:51];

for  n = 1:ns
    for iz =1:nz
        range_update = [];
        for ff = 1:nf
            
            gradientmap{n}(:,:,ff,iz) = abs(gradient(gradient(CO_imagchk{n}(:,:,ff,iz)'))');
            badphase_check = max(max(gradientmap{n}(:,:,ff,iz)));
           
            if badphase_check>1.5 %If it contains phase wrapping
                b = rng_goodphase(rng_goodphase~=ff);
                rng_goodphase = b;
            end
        end
        
    end
end

%correct the ranges to avoid the phase wrapping
for n = 1:ns
    s_CO_ryt{n} = s_CO_ryt{n}(:,:,rng_goodphase,:);
    s_OUT_ryt{n} = s_OUT_ryt{n}(:,:,rng_goodphase);
    
    %unimportant, only for visual interpretation
%     s_CO_ryt2{n} = s_CO_ryt2{n}(:,:,rng_goodphase,:);
    s_RO{n} = s_RO{n}(:,:,rng_goodphase);
    gradientmap{n} = gradientmap{n}(:,:,rng_goodphase);

end

nf = length(rng_goodphase);
ftable = ftable(rng_goodphase);
    
end

%% Preprocessing Reconstructions
%Sometimes there is a benefit from analyzing the raw data from the OUT
%reconstructions. These reconstructions are simply averaged signals across
%the frequency band. We can take a look at interesting regions in these
%reconstructions by applying apodization filters targetted towards those
%areas!
%The images generated right now are Summed signals, Gradient of summed
%signals (absolute along both x and y) and gradient squared of the summed
%signals.

if preprocess && comb_bornrytov
    for lol=1:length(ftable)
    deriv_OUT(:,:,lol) = abs(gradient(real(s_OUT_ryt{1}(:,:,lol,1)))) + abs(gradient(real(s_OUT_ryt{1}(:,:,lol,1))')');
  
    deriv2_OUT(:,:,lol) = abs(gradient(gradient(real(s_OUT_ryt{1}(:,:,lol,1))))) + abs(gradient(gradient(real(s_OUT_ryt{1}(:,:,lol,1))'))');
    end

figure;imagesc(sum(s_OUT_ryt{1},3));
 title('Summed Map of the OUT');
figure;imagesc(sum(deriv_OUT,3));
title('Summed Derivative Map of the OUT');
figure;imagesc(sum(deriv2_OUT,3));
title('Summed 2nd Derivative Map of the OUT');
elseif preprocess
        for lol=1:length(ftable)
    deriv_OUT(:,:,lol) = abs(gradient(real(s_OUT{1}(:,:,lol,1)))) + abs(gradient(real(s_OUT{1}(:,:,lol,1))')');
  
    deriv2_OUT(:,:,lol) = abs(gradient(gradient(real(s_OUT{1}(:,:,lol,1))))) + abs(gradient(gradient(real(s_OUT{1}(:,:,lol,1))'))');
    end

figure;imagesc(abs(sum(s_OUT{1},3)));
 title('Summed Map of the OUT');
figure;imagesc(sum(deriv_OUT,3));
title('Summed Derivative Map of the OUT');
figure;imagesc(sum(deriv2_OUT,3));
title('Summed 2nd Derivative Map of the OUT');
end

%% Part 1.2: Apodization Filtering
%--------------------------------------------------------------------------
% Important if there are large amounts of noise/discontinuities at the
% edges of the ROI. Typically left disabled in simulation results. Several
% filters are included, but using anything beyond the cosine filter
% requires specific adjustments.
%--------------------------------------------------------------------------
fftw('planner','exhaustive');

if  apod==1
    
    %cosine filter. This is the default apodization filter
    %----------------------------------------------------------------------
    w1 = cos(linspace(-pi/2, pi/2, nx));
    w2 = cos(linspace(-pi/2, pi/2, ny));
    cos_filt = w1'*w2;
 
    
   %other filters
   %-----------------------------------------------------------------------
    %There are two extra filters that you may wish to try, a gaussian
    %filter and the Happ-Genzel filter. They may need to be manually
    %adjusted in order to function corectly.
   apod_adj = 0.4;
   adj = apod_adj;
   stddev = nx/2*apod_adj;
    apod_filter = zeros(nx);
    for ix = 1:nx
        for iy = 1:ny
            
            %gaussian filter
            g_filt(ix,iy) = exp((-(ix-nx/2).^2)/(2*(stddev/2)^2))*exp((-(iy-ny/2).^2)/(2*(stddev/2)^2));
            
            %Happ-Genzel filter
            happ_genzel(ix,iy) = (0.54+0.46*cos(pi*(ix+nx/2)/(nx/2)))*(0.54+0.46*cos(pi*(iy+ny/2)/(ny/2)));
        end    
    end
    apod_filter2 = g_filt;
%     apod_filter = happ_genzel;
%   %-----------------------------------------------------------------------

tukeytol = 0.8;
tukey = tukeywin(nx,tukeytol)*tukeywin(nx,tukeytol)';
apod_filter = tukey;

ApodFilter = apod_filter;

% apod_filter = cos_filt;
else
    %if we dont want a gauss filter, just multiply everything by 1!
    apod_filter = ones(nx);
end

 %show the apodization filter
if show_apod
    fig=figure;
    dx = xstep;
    X = (-((nx-1)*dx)/2:dx:((nx-1)*dx)/2).*1000;
    set(fig, 'Position', [0, 450,400, 300]);
    imagesc(X,X,apod_filter);
    title('Apodization Filter');
        ystr = '{\ity} (mm)';
    xstr = '{\itx} (mm)';
        ylabel(ystr);
    xlabel(xstr); 
    colorbar;
        set(findall(fig,'-property','FontSize'),'FontSize',15)
        
        

end


if focusing

tuk_tol = 0.75;

range_x = focus_center(1)-focus_size/2:focus_center(1)+focus_size/2;
range_y = focus_center(2)-focus_size/2:focus_center(2)+focus_size/2;

tukfilter = tukeywin(focus_size+1,tuk_tol)*tukeywin(focus_size+1,tuk_tol)';
focus_filter = zeros(nx);

focus_filter(range_x,range_y) =tukfilter;

    if show_focus
        figure_2d(fliplr(rot90(focus_filter,-1)),10,0,1);
        title('Focusing Filter');
    end
else
    focus_filter = apod_filter;
end

%% Part 1.3: Normalization and the Fast Fourier Transform 
%--------------------------------------------------------------------------
% Compute the fourier transforms for the CO and OUT. We also normalize the
% signals to avoid biasing certain frequency values. The apodization is
% applied before the FFT
%--------------------------------------------------------------------------

% This section will remove poor SNR PSFs if necessary. It requires the
% PSF_Classification function to operate.
if clean_psf

for n = 1:ns 
    
    if comb_bornrytov
        psf_hold = PSF_classification(s_CO_ryt{n}, 'plot_CNR_SNR', 0,'CNR_thresholds', [0 0 0 1]);
    else
        psf_hold = PSF_classification(s_CO{n}(:,:,:,zplane), 'plot_CNR_SNR', 1,'CNR_thresholds', [0 0 0 2]);        
    end
     rng_goodphase = intersect(rng_goodphase,psf_hold.CNR_complex_mag_i);

end

%correct the ranges to avoid the phase wrapping
for n = 1:ns
    
    sc_CO{n} = sc_CO{n} (:,:,rng_goodphase,:);
    sc_OUT{n} = sc_OUT{n} (:,:,rng_goodphase,:);
    
    if comb_bornrytov
        s_CO_ryt{n} = s_CO_ryt{n}(:,:,rng_goodphase,:);
        s_OUT_ryt{n} = s_OUT_ryt{n}(:,:,rng_goodphase);
        gradientmap{n} = gradientmap{n}(:,:,rng_goodphase);
    end
    
    
    %unimportant, only for visual interpretation
    %     s_CO_ryt2{n} = s_CO_ryt2{n}(:,:,rng_goodphase,:);
    s_RO{n} = s_RO{n}(:,:,rng_goodphase);
end

nf = length(rng_goodphase);
ftable = ftable(rng_goodphase);

if nf==0
   disp('No Useable frequencies! Stopping reconstruction') 
    return;
end

end

%FFT is computed independently on each (x,y) image. 
% loop through each S-parameter
for n = 1:ns
    SF_CO{n} = zeros(nx,ny,nf,nz);
    SF_OUT{n} = zeros(nx,ny,nf);
    
    %loop through each frequency
    for fi = 1:nf
        %calculate the max value in the CO measurement at each frequency to
        %use in normalization
        maxi = max(max(max(abs(s_CO{n}(:,:,fi,:)))));
%         maxi = 1;
        
        if enable_rytov_real
            
            %Rytov OUT
            SF_OUT_ryt{n}(:,:,fi) = fft2(s_OUT_ryt{n}(:,:,fi)/maxi.*focus_filter);
            
            %Born OUT
            SF_OUT_born_real{n}(:,:,fi) = fft2(real(sc_OUT{n}(:,:,fi)/maxi).*focus_filter);
            SF_OUT_born_imag{n}(:,:,fi) = fft2(imag(sc_OUT{n}(:,:,fi)/maxi).*focus_filter);

            
            for iz = 1:nz
                CO_filter_norm{n}(:,:,fi,iz) = s_CO_ryt{n}(:,:,fi,iz).*apod_filter/maxi;% + (1-apod_filter)*mean(mean(s_CO_ryt{n}(1:10,1:10,fi,iz)/maxi));
                CO_filter_norm2{n}(:,:,fi,iz) = s_CO_ryt{n}(:,:,fi,iz).*apod_filter/maxi+ (1-apod_filter)*mean(mean(s_CO_ryt{n}(1:10,1:10,fi,iz)/maxi));

                
                %Rytov Real CO
                SF_CO_ryt_real{n}(:,:,fi,iz)=fft2(real(CO_filter_norm{n}(:,:,fi,iz)));
                
                %Rytov imaginary CO
                SF_CO_ryt_imag{n}(:,:,fi,iz)=fft2(imag(CO_filter_norm{n}(:,:,fi,iz)));
            
                %Born Real CO
                SF_CO_born_real{n}(:,:,fi,iz)=fft2(real(sc_CO{n}(:,:,fi,iz).*apod_filter/maxi));

                %Born imaginary CO
                SF_CO_born_imag{n}(:,:,fi,iz)=fft2(imag(sc_CO{n}(:,:,fi,iz).*apod_filter/maxi));
                
            end
            
        else
            
            %OUT FFT and normalization
            SF_OUT{n}(:,:,fi)=fft2(sc_OUT{n}(:,:,fi).*focus_filter/maxi);
            
            init_power{n}(fi) = sum(sum(abs(SF_OUT{n}(:,:,fi))));
            
            %CO FFT and normalization looped through each z plane
            for iz = 1:nz
                SF_CO{n}(:,:,fi,iz)=fft2((sc_CO{n}(:,:,fi,iz).*apod_filter/maxi));
            end
            
            if edot
                maxi = max(max(max(abs(S_res{n}(:,:,:,fi)))));
                maxi = 1;
                for iz = 1:nz
                    SF_CO{n}(:,:,fi,iz)=fft2((S_res{n}(:,:,iz,fi).*apod_filter/maxi));
                end
            end
        end
        
    end
   
end

%% Part 1.4: Solving the Inverse Problem: Least Squares Solution
%--------------------------------------------------------------------------
% Solve the QMH least squares problem. We go through each voxel and solve
% the least squares problem independently across the combinations of
% S-parameters, as well as combining the S-parameters together to find the
% minimum 2 norm solution. See equation 5.21
%
% Independent measurements are indentified by the suffix _ind
% Combined min 2 norm results are identified by the suffix _tot
% 
% Be careful when looking at indepedent data... it can generate strange
% values in simulated datasets! Especially when you have linear dependent
% CO planes
%
%
% Added in Revision 2: Tolerance determination through SVD
% By enabling set_pinv_tol = 1, the tolerance can be determined through the
% SVD function. The value for the tolerance cutoff will be selected by
% taking a diagonal slice of the SVD map, taking the derivative of this
% slice, and setting the tolerance to values below (1). This has been
% determined emperically, further study can be used to improve this. If you
% are curious to see what an SVD map is, uncomment the appropriate section
% in the psuedoinverse section of the code.
%
%--------------------------------------------------------------------------
if set_tol
    
    %obtain the diagonal of the SVD; kxy = (kx,kx). This captures the largest kx/ky
    %responses, capturing the typically smallest SVD value.
    for kxy = 1:nx
        
        A_ind = [];
        A_tot = [];
                
        for n = 1:ns
            
            if comb_bornrytov
                %Rytov Real
                A_ind = [reshape(SF_CO_ryt_real{n}(kxy,kxy,:,:),[nf,nz]) -reshape(SF_CO_ryt_imag{n}(kxy,kxy,:,:),[nf,nz])];
                %Born real
                A_ind  = [A_ind;reshape(SF_CO_born_real{n}(kxy,kxy,:,:),[nf,nz]) -reshape(SF_CO_born_imag{n}(kxy,kxy,:,:),[nf,nz])];
%                 %Born imaginary
                A_ind = [A_ind;reshape(SF_CO_born_imag{n}(kxy,kxy,:,:),[nf,nz]) reshape(SF_CO_born_real{n}(kxy,kxy,:,:),[nf,nz])];

            elseif enable_rytov_real
                %Rytov Real
                A_ind = [reshape(SF_CO_ryt_real{n}(kxy,kxy,:,:),[nf,nz]) -reshape(SF_CO_ryt_imag{n}(kxy,kxy,:,:),[nf,nz])];
                
            else
                A_ind = reshape(SF_CO{n}(kxy,kxy,:,:),[nf,nz]);
            end
            
            %SVD diagonal captured.
            svd_diagonal_ind(kxy,:,n) = svd(A_ind);
            
            %The S-parameters are stacked to process the result processed
            %using all captured data.
            A_tot = [A_tot;A_ind];            
        end
        
        svd_diagonal_tot(kxy,:) = svd(A_tot);
        
    end
    
    %right now this selection is arbitrary. More accurate approaches based on
    %finding the regions where the signal is essentially random and removing
    %those responses is necessary.
    tol = 5*min(min(svd_diagonal_tot));
    
else
    tol = 10^-15; %default tolerance
end
% tol = 10^0.4;

%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
% UNDER DEVELOPMENT: If this is enabled, each frequency measurement will be
% filtered according to that frequencies maximum possible kz

if freq_filter
    
    for n = 1:ns
        for ff= 1:nf
            
            lambda = 2.99792458e8/ftable(ff)/sqrt(real(eps_RO));
            
            kmax = 2*2*pi/lambda;
            kx = 2*pi/xstep;
            dkx = kx/nx;
            kmax_steps = kmax/dkx*0.1;
            
            fs = 2*pi/dx;
            df = fs/(nx-1);
            
            %set the orde of the Butterworth filter
            order=8;
            bfilt = ones(nx,ny);
            %BUTTERWORTH FILTER
            dev = kmax_steps;
            for i = 1:nx
                for j =1:ny
                    chg = ((i-((nx-1)/2+1))^2 + (j-((ny-1)/2+1))^2)^.5;
                    % Create Butterworth filter.
                    bfilt(i,j)= 1/(1 + (chg/dev)^(2*order));
                end
            end
            if n == 1
%                 figure;
%                 subplot(2,1,1)
%                 imagesc(abs(sc_OUT{1}(:,:,ff)))
%                 subplot(2,1,2)
%                 imagesc(abs(fftshift(SF_OUT{1}(:,:,ff))))
%                 viscircles([(nx-1)/2+1,(nx-1)/2+1],dev)
%                 %             figure;imagesc(bfilt)
%                 title(ff)
            end
            
            freq_filt(:,:,ff) = fftshift(bfilt);
            
        end
        
        if enable_rytov_real
            
            SF_OUT_born_real{n} = SF_OUT_born_real{n}.*freq_filt;
            SF_OUT_born_imag{n} = SF_OUT_born_imag{n}.*freq_filt;
            
            SF_OUT_ryt{n} = SF_OUT_ryt{n}.*freq_filt;
            
            for iz = 1:nz
                SF_CO_born_real{n}(:,:,:,iz) = SF_CO_born_real{n}(:,:,:,iz).*freq_filt;
                SF_CO_born_imag{n}(:,:,:,iz) = SF_CO_born_imag{n}(:,:,:,iz).*freq_filt;
                
                SF_CO_ryt_real{n}(:,:,:,iz) = SF_CO_ryt_real{n}(:,:,:,iz).*freq_filt;
                SF_CO_ryt_imag{n}(:,:,:,iz) = SF_CO_ryt_imag{n}(:,:,:,iz).*freq_filt;              
            end
        else
            
            SF_OUT{n} = SF_OUT{n}.*freq_filt;
            for iz = 1:nz
            SF_CO{n}(:,:,:,iz) = SF_CO{n}(:,:,:,iz).*freq_filt;
            end
        end
        
        
    end
end
 disp("Pre Inversion")
toc
%--------------------------------------------------------------------------
%--------------------------------------------------------------------------
%--------------------------------------------------------------------------

%% Matrixless inversion UNDER DEVELOPMENT
%=====================================================================
%=====================================================================
%=====================================================================
%First calculate the desired range resolution

if matrixless
    
    %Define relevant parameters
    z_list = [0.01 0 -0.01];
    z_list = [0];
    nz = length(z_list);
    omega_list = 2*pi*F';
    dw = omega_list(2)-omega_list(1);
    
    % calculate kx, ky
    fs = 2*pi/xstep; %frequency sampling in rad/s
    df = fs/(nx-1); %sampling step in rad per second
    %this is the range in the kx and ky
    kx_list = (-fs/2:df:fs/2);
    ky_list = (-fs/2:df:fs/2);
    delta_eps = (eps_CO-eps_RO);
    
    multi_factor = delta_eps*vol_sp/(xstep^2)*...
        4/(2*pi*cb^2);
    
    k0 = omega_list/cb;
    
    
    %generate an image filter at each frequency to remove spikes in the PSF
    %division portion of the equation
    
    
        %set the orde of the Butterworth filter
        order=8;
        bfilt = ones(nx,ny,nf);
        %BUTTERWORTH FILTER
    for ff = 1:nf
        kmax = 2*k0(ff)/df;
        dev = kmax-kmax*0.2;
        

        for i = 1:nx
            for j =1:ny
                chg = ((i-((nx-1)/2+1))^2 + (j-((ny-1)/2+1))^2)^.5;
                % Create Butterworth filter.
                bfilt(i,j,ff)= 1/(1 + (chg/dev)^(2*order));
            end
        end
        bfilt(:,:,ff) = bfilt(:,:,ff)/sum(sum(bfilt(:,:,ff)))*nx*ny;
    end
    
    %Transform the frequency responses of OUT and CO to the correct k-space
    %orientation (where kx,ky = 0 is in the center)
    
    for s = 1:ns
        for ff = 1:nf
            SF_OUT{s}(:,:,ff) = fftshift(SF_OUT{s}(:,:,ff));
            for zz = 1:nz
                SF_CO{s}(:,:,ff,zz) = fftshift(SF_CO{s}(:,:,ff,zz));
            end
        end
    end
  
                
for kx = 1:nx
    for ky = 1:ny
        data_div_mat(kx,ky,:) = squeeze(SF_OUT{1}(kx,ky,:))./squeeze(SF_CO{1}(kx,ky,:,1));
        for zz = 1:length(z_list)
                kz_map(kx,ky,:) = sqrt((2*omega_list/cb).^2-kx_list(kx)^2-ky_list(ky)^2);
                kz = sqrt((2*omega_list/cb).^2-kx_list(kx)^2-ky_list(ky)^2);

                
                OUT_div_PSF(kx,ky,:) = (squeeze(SF_OUT{1}(kx,ky,:))./squeeze(SF_CO{1}(kx,ky,:,1))).*squeeze(bfilt(kx,ky,:));

%                 kz_map(kx,ky,imag(kz)~=0) = 0;
%                 kz(imag(kz)~=0) = 0;
                
                Fout_tot(kx,ky,zz) = delta_eps*vol_sp/(xstep^2)*...
                    4/(2*pi*cb^2)*...
                    sum(squeeze(OUT_div_PSF(kx,ky,:)).*...
                    exp(1i*kz*z_list(zz))./kz.*omega_list*dw);
                
%                 Fout_tot(kx,ky,zz) = delta_eps*vol_sp/(xstep^2)*...
%                     4/(2*pi*cb^2)*...
%                     sum(squeeze(SF_OUT{1}(kx,ky,:))./squeeze(SF_CO{1}(kx,ky,:,2)).*...
%                     exp(1i*kz*z_list(zz)).*omega_list*dw);
                
                
                Fout_inter(kx,ky,zz,:) = delta_eps*vol_sp/(xstep^2)*...
                    4/(2*pi*cb^2)*...
                    (squeeze(OUT_div_PSF(kx,ky,:)).*...
                    exp(1i*kz*z_list(zz))./kz.*omega_list*dw);
                
                kz_stuff(kx,ky,zz) = sum(exp(1i*kz*z_list(zz))./kz.*omega_list*dw)/(2*pi*cb^2);
                kz_stuff_plot(kx,ky,:,zz) = (exp(1i*kz*z_list(zz))./kz.*omega_list*dw)/(2*pi*cb^2);

                kz_exp(kx,ky,zz) = sum(exp(1i*kz*z_list(zz)));
                
             
        end
    end
end

        for zz = 1:nz
            Fout_tot(:,:,zz) = ifftshift(Fout_tot(:,:,zz));
        end
        Fout_ind = Fout_tot;
    
else
%% Typical Inversion

% at every position (kx,ky) the inversion is performed
for kx = 1:nx
    for ky = 1:ny
        
        A_ind = [];
        b_ind = [];
        A_tot = [];
        b_tot = [];
        
        if enable_rytov_real
            A_ind_comb = [];
            b_ind_comb = [];
            
            A_tot_comb = [];
            b_tot_comb = [];
        end

        %This loop process each S-parameter independently
        for n = 1:ns

            A_ind = reshape(SF_CO{n}(kx,ky,:,:),[nf,nz]);
            b_ind = reshape(SF_OUT{n}(kx,ky,:),[nf, 1]);
            
            %The S-parameters are stacked to process the result processed
            %using all captured data. 
            
            A_tot = [A_tot;A_ind];
            b_tot = [b_tot;b_ind];
 
            %SVD map to reduce high-frequency noise on reconstruction
            svd_map_ind(kx,ky,:,n) = svd(A_ind);
            
            %Perform inversion on ONLY 1 S-parameter
            Fout_ind(kx,ky,:,n) = pinv(A_ind)*(b_ind);
            
            if enable_rytov_real
                %Rytov real
                A_ind_comb = [reshape(SF_CO_ryt_real{n}(kx,ky,:,:),[nf,nz]) -reshape(SF_CO_ryt_imag{n}(kx,ky,:,:),[nf,nz])];
                
                %Rytov on top
                b_ind_comb = [reshape(SF_OUT_ryt{n}(kx,ky,:),[nf,1])];
                
                if comb_bornrytov
                    
                    %Born real
                    A_ind_comb  = [A_ind_comb;reshape(SF_CO_born_real{n}(kx,ky,:,:),[nf,nz]) -reshape(SF_CO_born_imag{n}(kx,ky,:,:),[nf,nz])];
                    %                 %Born imaginary
                    A_ind_comb = [A_ind_comb;reshape(SF_CO_born_imag{n}(kx,ky,:,:),[nf,nz]) reshape(SF_CO_born_real{n}(kx,ky,:,:),[nf,nz])];
                    
                    
                    %Born real
                    b_ind_comb = [b_ind_comb;reshape(SF_OUT_born_real{n}(kx,ky,:),[nf,1])];
                    %                 %Born imaginary
                    b_ind_comb = [b_ind_comb;reshape(SF_OUT_born_imag{n}(kx,ky,:),[nf,1])];
                    
                end
                Fout_ind_comb(kx,ky,:,n) =  pinv(A_ind_comb,tol)*b_ind_comb;
                
                A_tot_comb = [A_tot_comb;A_ind_comb];
                b_tot_comb = [b_tot_comb;b_ind_comb];
                
            end
            
        end
   
        %SVD map to reduce high-frequency noise on reconstruction
%         svd_map_tot(kx,ky,:) = svd(A_ind);

        
        %Perform inversion on the combined S-parameters
        Fout_tot(kx,ky,:) = pinv(A_tot,tol)*b_tot;
        svd_map_tot(kx,ky,:) = svd(A_tot);

        if enable_rytov_real
            Fout_tot_comb(kx,ky,:) = pinv(A_tot_comb,tol)*b_tot_comb;
            svd_map_totcomb(kx,ky,:) = svd(A_tot_comb);
            svd_map_outcomb(kx,ky,:) = svd(b_tot_comb);

        end
        
    end
end

  disp("Post Inversion")
toc


end

if show_svd
    if enable_rytov_real
        figure;imagesc(log10(svd_map_outcomb(:,:,1)));
        title('SVD - No Thresholding');
        svd_map_outcomb((svd_map_outcomb(:,:,1))<tol) = 0;
        figure;imagesc(log10(svd_map_outcomb(:,:,1)));
        title('SVD - With Thresholding');
        
    else
        figure;imagesc(log10(svd_map_tot(:,:,1)));
        title('SVD - No Thresholding');
        svd_map_tot((svd_map_tot(:,:,1))<tol) = 0;
        figure;imagesc(log10(svd_map_tot(:,:,1)));
        title('SVD - With Thresholding');
    end
    
end

% figure;imagesc(abs(Fout_tot(:,:,3)));

%% Part 1.5: Post-inversion Fourier Low Pass filtering
%--------------------------------------------------------------------------
% Filtering has been show to be a vital part of the process of quantitative
% and qualititative holography, especially when there is oversampling
% involved. Now we will generate a Butterworth filter to remove high
% frequency noise reponses. See [Tajik, JPIER 2017] for more
% information.
% The filter sets the limit of the fourier space equal to the limit as
% stated in section 5.3 of Introduction of Microwave Imaging
%--------------------------------------------------------------------------

% for ii = 1

lambda_min = c0/sqrt(abs((eps_RO)))/ftable(nf);
    %lambda_min = c0/sqrt(abs((eps_RO)))/ftable(3);
    
    %Calculate the maximum angle the antenna can see. This assumes that the
    %directivity of the antenna does not affect the maximum angle, which is
    %typically not the case in experimental reconstructions
    dx = xstep;
    %calculate maximum kmax, and determine is location in the physical
    %FFT images
    %kmax = 4*pi*sin(alpha)/lambda_min;
    kmax = 2*2*pi/lambda_min*filter_scale;
    alpha= asind(filter_scale);
    kx = 2*pi/xstep;
    dkx = kx/nx;
    kmax_steps = kmax/dkx;
    
    fs = 2*pi/dx;
    df = fs/(nx-1);

    %this is the range in the kx and ky
    kxx = (-fs/2:df:fs/2);
    
    

    
if enable_lpf == 1
   
    %calculate minimum wavelength
    
    
    
    dev = kmax_steps;
    
    %set the orde of the Butterworth filter
    order=8;
    bfilt = ones(nx,ny);
    %BUTTERWORTH FILTER
    for i = 1:nx
        for j =1:ny
            chg = ((i-((nx-1)/2+1))^2 + (j-((ny-1)/2+1))^2)^.5;
            % Create Butterworth filter.
            bfilt(i,j)= 1/(1 + (chg/dev)^(2*order));
        end
    end
    
    %LowPassFilter = lpf;
    %Shift the filter to align with the FFT versions of the CO and OUT
    lpf = fftshift(bfilt);
    
    %Any values below 0.01 are automatically set to 0. This reduces
    %computational costs
    lpf = lpf.*(lpf>0.01);
    
        %TESTING: Make the LPF scale so that the overal summation of the filter
    %remains the same as if there was no window (i.e. try to move the
    %overal signal power from the erroneous high frequnecy responses to the
    %lower frequency values
    
     lpf = lpf/sum(sum((lpf.^2)))*nx*ny;
else
    %if we dont want to filter, just multiply everything by 1!
    lpf = ones(nx);
end


 %show the filter
if show_lpf
    fig = figure;
    dx = xstep;
    
    set(fig, 'Position', [450, 450,400, 300]);
    imagesc(kxx,kxx,ifftshift(lpf));
    hold on;
    viscircles([0,0],kmax)
    ystr = '{\itky} (rad/m)';
    xstr = '{\itkx} (rad/m)';
    ylabel(ystr);
    xlabel(xstr); 
    colorbar;
       title('Low Pass Filter');
    set(findall(fig,'-property','FontSize'),'FontSize',15)
    
end

if enable_rytov_real
   Fout_tot = Fout_tot_comb; 
end
    fig = figure;
    dx = xstep;
    
    set(fig, 'Position', [900, 450,400, 300]);
    imagesc(kxx,kxx,fftshift(abs(Fout_tot(:,:,1))));
    hold on;
    viscircles([0,0],kmax)
    ystr = '{\itky} (rad/m)';
    xstr = '{\itkx} (rad/m)';
    ylabel(ystr);
    xlabel(xstr); 
    colorbar;
       title('Permittivity Map in Fourier Domain, Mag.');
    set(findall(fig,'-property','FontSize'),'FontSize',15)

%% Part 1.7: Final image adjustments (smoothing)
%--------------------------------------------------------------------------
%In case you want to make prettier images, you can enable this simple
%smoothing strategy.
%--------------------------------------------------------------------------


%this factor mulitplies with the reconstruction results. See equation 5.22
chg_eps = (vol_sp/(xstep^2*zstep))*(eps_CO-eps_RO);

% air_chg = (1+2*eps_RO)/(eps_CO+2*eps_RO);
% chg_eps = (vol_sp/(xstep^2*zstep))*(eps_CO-eps_RO)*air_chg;

epsout_tot = [];

for iz = 1:nz
    
    test = lpf.*Fout_tot(:,:,iz);
%     in_power = max(init_power{2});
%     out_power = sum(sum(abs(Fout_tot)));
    
    %If we have combined Born and Rytov, we need to reorganize the
    %matrix so that we have a complex number at each index, instead of
    %real numbers and imaginary numbers separated. Note that the real
    %and imaginary numbers are stacked on top of eachother, hence the
    %(:,:,iz,n) for real numbers and (:,:,iz+nz,n) for imaginary
    if enable_rytov_real
        fout_tot(:,:,iz) = (fftshift(ifft2(lpf.*(Fout_tot_comb(:,:,iz)))))+1i*(fftshift(ifft2(lpf.*(Fout_tot_comb(:,:,iz+nz)))));
        
        fout_tot_real(:,:,iz) = (fftshift(ifft2(lpf.*(Fout_tot_comb(:,:,iz)))));
        fout_tot_imag(:,:,iz) = 1i*(fftshift(ifft2(lpf.*(Fout_tot_comb(:,:,iz+nz)))));
    else
        fout_tot(:,:,iz) = (fftshift(ifft2(lpf.*Fout_tot(:,:,iz))));
%         fout_tot(:,:,iz) = (fftshift(ifft2(lpf.*Fout_tot(:,:,iz)*in_power/out_power)));
    end
    
    
    % Apply the multiplicative processing to obtain quantitative
    % information. See equation 5.22
    epsout_tot(:,:,iz) = chg_eps*fout_tot(:,:,iz) + eps_RO;
    
    
    % The above is repeated for each independent S-parameter
    for n = 1:ns
         
        if enable_rytov_real
            fout_ind(:,:,iz,n) = (fftshift(ifft2(lpf.*(Fout_ind_comb(:,:,iz,n)))))+1i*(fftshift(ifft2(lpf.*(Fout_ind_comb(:,:,iz+nz,n)))));;
        else
            fout_ind(:,:,iz,n) = (fftshift(ifft2(lpf.*(Fout_ind(:,:,iz,n)))));
            
        end
        
        %equation 5.22
        epsout_ind(:,:,iz,n) = chg_eps*fout_ind(:,:,iz,n) + eps_RO;
        
    end
    
end


%Smoothing reduces the impact of sharp variations in the image
%reconstruction. This can be helpful if you are attempting to resolve
%large objects which appear muddied by noise
if smoothing == 1
    
    %resample creates rippling effects. Reduce these by expanding the
    %image, resampling, then cutting off the edge of the image.
    ripple_pad = round(nx*0.5);
    new_nx = nx + 2*ripple_pad;
    holder_tot = padarray(epsout_tot,[ripple_pad ripple_pad],'replicate','both');
    holder_ind = padarray(epsout_ind,[ripple_pad ripple_pad],'replicate','both');
    
    epsout_tot = [];
    epsout_ind = [];
    for iz = 1:nz
        %To improve visual quality, the data is resampled at a higher
        %number of x,y samples. This is then filtered.
        p = 4;
        q = 1;
        
        %has to be done one dimension at a time
        
        first_dim = resample(holder_tot(:,:,iz),p,q);
        both_dim = resample(first_dim',p,q)';      
        
        ripple_rng = ripple_pad*p/q:(new_nx-ripple_pad)*p/q;
        
        resamp_removed = both_dim(ripple_rng,ripple_rng);
        
        %This is the filter. essentially it averages the pixels around the
        %selected pixel. This can be modified if desired.
        h = 1/4*[1 1;1 1];
        h = 1/9*[1 1 1;1 1 1;1 1 1];
        epsout_tot(:,:,iz) = imfilter(resamp_removed,h,'replicate');

        
        for n = 1:ns
            %To improve visual quality, the data is resampled at a higher
            %number of x,y samples. This is then filtered.
            p = 4;
            q = 1;
            
            %has to be done one dimension at a time
            first_dim = resample(holder_ind(:,:,iz,n),p,q);
            both_dim = resample(first_dim',p,q)';
            
            %This is the filter. essentially it averages the pixels around the
            %selected pixel. This can be modified if desired.
            h = 1/4*[1 1;1 1];
            
            epsout_ind(:,:,iz,n) = imfilter(both_dim,h,'replicate');
        end
    end
end

    
%% Part 1.6: Multiplicative factoring and optimization
%--------------------------------------------------------------------------
%This is the multiplication factor that produces the quantitative
%information. In the case of nonphysical values, the optimization can be
%enabled.
%
%Note: You must have CVX installed for this to work. If you do no have it
%installed, the optimization will error out.
%--------------------------------------------------------------------------

%Here begins the optimization that is described in: TAJIK EUCAP 2018

if correct_nonphys

%CVX does not allow the matlab FFT function. This means we are forced into
%using the DFT to solve the problem.
dftmat = dftmtx(nx);


%Combine the data structures for simplified code. If the user only cares
%about the combined dataset, don't solve the independent dataset. The vice
%versa scenario is also true. 

all_eps = zeros(nx,ny,nz,ns+1);
all_rho = zeros(nx,ny,nz,ns+1);

samp_rng = 0;

if ~close_ind
    
    samp_rng = 1:ns;
    
    all_rho(:,:,:,samp_rng) = fout_ind;
    all_eps(:,:,:,samp_rng) = chg_eps.*fout_ind + eps_RO;
    
end

if ~close_comb
    
    if samp_rng(1) == 1;
        samp_rng = 1:ns+1;
    else
        samp_rng = 1;
    end
    
    all_rho(:,:,:,samp_rng(end)) = fout_tot;
    all_eps(:,:,:,samp_rng(end)) = chg_eps.*fout_tot + eps_RO;

end

%This is the calculated background permittivity
epsRO = ones(nx*ny*nz,1)*eps_RO;

%Calculate the original err of the unoptimized QMH forward model
%If it is the combined dataset, form the combined matrices

%rho_psf_int = zeros(nx,ny,nf,samp_rng);
orig_err = zeros(nf,length(samp_rng));

for ff = 1:nf
    for samps = samp_rng
        rho_psf_int = 0;
       % hold = zeros(nx,ny);
        for zz = 1:nz
        
            hold_rho = all_rho(:,:,zz,samps);
            hold_PSF = SF_CO{samps}(:,:,ff,zz);
        %this is the integral part of equation (19) in EUCAP paper
         rho_psf_int = rho_psf_int + hold_rho(lpf>0).*hold_PSF(lpf>0)*zstep;
    
        end
        
        hold_OUT = SF_OUT{samps}(:,:,ff);
        orig_err(ff,samps) = norm(hold_OUT(lpf>0)-rho_psf_int);
        
     end
end


for samps = samp_rng
             
%CVX solves the problem faster if it is reformated into a single vector.
%Thus, the permittivity and the reflectivity function are reformated

%This is the calculated permittivity using the unoptimized QMH
epsConst = reshape(all_eps(:,:,:,samps),[nx*ny*nz,1]);

%This is the calculated reflectivity constant using the unoptimized QMH
rhoConst = reshape(all_rho(:,:,:,samps),[nx*ny*nz,1]);



%Initialize CVX
 cvx_begin
     %The MOSEK solver seems to be the best at solving this problem
     cvx_solver Mosek %Mosek, SDPT3 sedumi gurobi
     
     %The optimized variable that will change the reflectivity function
     variable x complex
    
     %The objective function (see EUCAP 2018)
     minimize(norm(epsConst-(x*rhoConst+epsRO),1))
     
     subject to
     
     
     real(x*rhoConst+epsRO)>=1; %real constraint
     imag(x*rhoConst+epsRO)<=0; %imaginary constraint


     %The optimized permittivity
     epsOUT = reshape(x.*rhoConst+epsRO,[nx,ny,nz]);
            
     for zz = 1:nz
         rhoBAR(:,:,zz) = dftmat*(xstep^3*(epsOUT(:,:,zz)-eps_RO)/(eps_CO*vol_sp))*dftmat';
     end
     
     %Calculate the error in the new solution at each signal frequency. 
     %This strategy only solves the problem in the locations where values
     %are not 0, which speeds up the process of optimization
     for ff = 1:nf
         
         rho_psf = 0;
          % The optimized reflectivity function. It is transformed into the
        % frequency domain using a discrete Fourier transform.
        for zz = 1:nz
            hold_rho = all_rho(:,:,zz,samps);
            hold_PSF = SF_CO{samps}(:,:,ff,zz);
            rho_psf = rho_psf + hold_rho(lpf>0).*hold_PSF(lpf>0);
        end
         
        hold_OUT = SF_OUT{samps}(:,:,ff);
        %for nn = 1:length(rows)                
        err = hold_OUT(lpf>0) - rho_psf;
        %end
        %find the error!
        norm(err)-orig_err(ff,samps)<=0;
     end
% %          
                
%          iff = 1
%            norm(norm((SF_OUT{1}(:,:,iff) - rhoBAR.*SF_CO{1}(:,:,iff)))) - norm(norm(SF_OUT{1}(:,:,iff) - Fout_tot.*SF_CO{1}(:,:,iff)))<=0;
%          
         
         %holder<=0;
         
         %norm(norm((SF_OUT{1} - rhoBAR.*SF_CO{1}))) - norm(norm(SF_OUT{1} - Fout_tot.*SF_CO{1}))<=0
       
%         for ix = 1:nx f
%            for iy = 1:ny
%                con = con + (epsOUT(ix,iy)-eps_RO)/chg_eps.*circshift(s_CO{1}(:,:,fu),[-(nx-1)/2+ix,-(ny-1)/2+iy]);
%            end
%         end
%          
%        norm(epsOUT - reshape(con,[nx*ny,1]),2)==0;
%          
 cvx_end

%norm(epsConst - (chg_eps*rhoConst+eps_RO));
 
disp('end')

%epsout_tot(:,:) = x*fout_ind(:,:,1,2)+eps_RO;   


%If both the separate and combined datasets are being computed
if length(samp_rng) == ns+1
    if samps <= ns
        epsout_ind(:,:,:,samps) = x*fout_ind(:,:,:,samps)+eps_RO;
    else
        epsout_tot(:,:,:) =  x*fout_tot(:,:,:)+eps_RO;
    end
    
% if only the independent datasets are being computed
elseif ~close_ind
    epsout_ind(:,:,:,samps) = x*fout_ind(:,:,:,samps)+eps_RO;

% if only the combined dataset is being computer
else
    epsout_tot(:,:,:) =  x*fout_tot(:,:,:)+eps_RO;   
end
    


    end


end
    


%% UNDER DEVELOPMENT: Isolating Permittivity Invalid Regions 
% go through the image and identify regions that are outside of 10% of the
% permittivies and set everything else to be 0.

if combine_COs
co_distlim = 0.2;

for iz = 1:nz
    for ix = 1:nx
        for iy = 1:ny
            if ~(real(epsout_tot(ix,iy,iz))<real(eps_CO)+(abs(eps_CO)*co_distlim) && real(epsout_tot(ix,iy,iz))>real(eps_CO)-(abs(eps_CO)*co_distlim) || real(epsout_tot(ix,iy,iz))<real(eps_RO)+(abs(eps_RO)*co_distlim) && real(epsout_tot(ix,iy,iz))>real(eps_RO)-(abs(eps_RO)*co_distlim))
                 epsout_tot(ix,iy,iz) = NaN;
            end
            if ~(imag(epsout_tot(ix,iy,iz))<imag(eps_CO)+(abs(eps_CO)*co_distlim) && imag(epsout_tot(ix,iy,iz))>imag(eps_CO)-(abs(eps_CO)*co_distlim) || imag(epsout_tot(ix,iy,iz))<imag(eps_RO)+(abs(eps_RO)*co_distlim) && imag(epsout_tot(ix,iy,iz))>imag(eps_RO)-(abs(eps_RO)*co_distlim))
                 epsout_tot(ix,iy,iz) = NaN;
            end
        end
    end
end
end
%% Part 1.7: Image Generation
%Final images are generated here.
 

epsout_tot = (rot90(epsout_tot));
zshift = 0;
for iz = 1:nz
    
    for n = 1:ns

        
        maintitle = sprintf('S-param = %d, Z-plane = %.02f %s',n,z(iz),units);
        filename = sprintf('S%d_z%.03f.fig',n,z(iz));
        
        
        fig = figure;
        set(fig,'name',maintitle,'numbertitle','off')
        
        %First subplot is the magnitude of the whole image
        subplot(1,3,1);
        imagesc(abs(epsout_ind(:,:,iz,n)));
        title('Magnitude');
        colorbar;
%         caxis(mag_ax) 
        set(gca,'ydir','normal');

        %Second subplot is the real part of the image
        subplot(1,3,2);
        imagesc(real(epsout_ind(:,:,iz,n)));       
        title('Real');
        set(gca,'ydir','normal');
        colorbar;
%         caxis(re_ax) 

        %third subplot is the imaginary part of the image
        subplot(1,3,3);
        imagesc(imag(epsout_ind(:,:,iz,n)));
        title('Imaginary');
        colorbar;
%         caxis(im_ax) 
        set(gca,'ydir','normal');
        
        set(findall(fig,'-property','FontSize'),'FontSize',15)
        set(fig, 'Position', [100, 100, 1200, 300]);

        % % enable this if you want to save the figures
        if save_ind savefig(filename); end
        if close_ind close(fig); end
        
    end
    
    %same as above: This image being printed is the total combined image
    %using all the s-parameters provided.
    maintitle = sprintf('Reconstructed Image, Z-plane = %.03f %s',z(iz),units);
    filename = sprintf('%s_z%.03f.fig','S_Combo_',z(iz));

    fig = figure;
    set(fig,'name',maintitle,'numbertitle','off')

    subplot(1,3,1);
    imagesc(abs((epsout_tot(:,:,iz))));
    title('Magnitude');
    colorbar;
    set(gca,'ydir','normal');
%     caxis(mag_ax)
    
    
    subplot(1,3,2);
    imagesc(real((epsout_tot(:,:,iz))));
    title('Real');
    set(gca,'ydir','normal');
    colorbar;
%     caxis(re_ax)
    
    
    subplot(1,3,3);
    imagesc(imag((epsout_tot(:,:,iz))));
    title('Imaginary');
    colorbar;
    set(gca,'ydir','normal');
%     caxis(im_ax)
    
    
    set(findall(fig,'-property','FontSize'),'FontSize',15)
    set(fig, 'Position', [50+zshift, 75, 1200, 265]);

    if save_comb savefig(filename); end
    if close_comb close(fig); end
    
    zshift = zshift +25;
end

%%
disp(" ")

disp("QMH Execution Complete!")

disp("=======================")

if enable_rytov_real && ~comb_bornrytov
    enable_rytov = 1;
    disp('Computed Rytov-Real only Approach')
elseif comb_bornrytov
    enable_rytov = 0;
    enable_rytov_real = 1;
    disp('Computed Combined Born/Rytov Approach')
elseif enable_rytov
    disp('Computed Rytov Approach - Be Aware of Phase Wrapping Issues!')  
else
    disp('Computed Born Approach')
end

if apod
    disp("Apodization Enabled")
    disp(" - Tukey Filter Tolerance set to: " + tukeytol)
else
    disp("No Apodization Applied")
end

if enable_lpf
    disp("Low-pass Filtering Enabled")
    disp(" - Alpha set to " + alpha + " degrees")
else
    disp("No Low-pass Filtering Applied")
end

disp("Measurement Parameters")
disp(" - Spatial Step Size = "+ xstep + " m")
disp(" - Frequency Range = "+ ftable(1)/1e9 + " GHz to "+ftable(end)/1e9+ " GHz in " + (ftable(2)-ftable(1))/1e9 + " GHz increments" )
